/*
 * Copyright 2015 Google Inc. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

table.google-visualization-table-table, table.dataframe {
  font-family: inherit;
  font-size: smaller;
}
tr.gchart-table-row {
}
tr.gchart-table-headerrow, table.dataframe thead th {
  font-weight: bold;
  background-color: #e7e7e7;
}
tr.gchart-table-oddrow, table.dataframe tr:nth-child(odd) {
  background-color: #f7f7f7;
}
tr.gchart-table-selectedTableRow {
  background-color: #e3f2fd;
}
tr.gchart-table-hoverrow, table.dataframe tr:hover {
  background-color: #bbdefb;
}
td.gchart-table-cell, table.dataframe td {
  border: solid 1px #cfcfcf;
}
td.gchart-table-rownumcell, table.dataframe tr th {
  border: solid 1px #cfcfcf;
  color: #999;
}
th.gchart-table-headercell, table.dataframe th {
  border: solid 1px #cfcfcf;
}
div.bqgc {
  display: flex;
  justify-content: center;
}
div.bqgc img {
  max-width: none; // Fix the conflict with maps and Bootstrap that messes up zoom controls.
}
.gchart-slider {
  width: 80%;
  float: left;
}
.gchart-slider-value {
  text-align: center;
  float: left;
  width: 20%;
}
.gchart-control {
  padding-top: 10px;
  padding-bottom: 10px;
}
.gchart-controls {
  font-size: 14px;
  color: #333333;	 
  background: #f4f4f4;
  padding: 10px;
  width: 180px;
  float: left;
}
.bqgc {
  padding: 0;
  max-width: 100%;
}
.bqgc-controlled {
  display: flex;
  flex-direction: row;
  justify-content:space-between;
}
.bqgc-container {
  display: block;
}
.bqgc-ml-metrics {
  display: flex;
  flex-direction: row;
  justify-content:left;
}
