/*
 * Copyright 2015 Google Inc. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

table.bqsv {
  font-family: inherit;
  font-size: smaller;
}
table.bqsv th, table.bqsv td {
  border: solid 1px #cfcfcf;
}
th.bqsv_expanded, th.bqsv_collapsed {
  background-color: #f7f7f7;
}
th.bqsv_colheader {
  font-weight: bold;
  background-color: #e7e7e7;
}
tbody.bqsv_hidden {
  display: none;
}
th.bqsv_expanded:before {
  content: '\25be  '
}
th.bqsv_collapsed:before {
  content: '\25b8  '
}
